#!/usr/bin/env python
# -*- coding: utf-8 -*-
import datetime
import hashlib, re
from bson.json_util import dumps
from bson.json_util import loads
from devdoo.validate import Validate


class DataRest():
    def __init__(self, message, custom_service):

        # TODO:: Implementar configuração de limites definido no console para o serviço
        self.__max_limit = 200
        self.__default_limit = 100

        self.__validate = Validate(custom_service)
        self.__filter = None
        self.__data = {}
        self.decode(message)

    def action(self, action):
        self.__action = action

    def cursor(self, action):
        self.__cursor = cursor

    def data(self, data):
        self.__data = data
        # Registra o checksum do documento
        self.checksum(data)

    def fields(self, fields):
        self.__fields = fields

    # --------------------------------
    # filter
    # --------------------------------
    # Filtra dados na base de dados conforme parametros definidos
    #
    # { price: { $not: { $gt: 1.99 } }
    # filter=price.not.gt=1.99
    #
    def filter(self):
        def dot_query_operators(params):
            filter_params = []
            result = {}

            params_split = params.split(";")
            if len(params_split) > 0:
                for item_param in params_split:
                    if item_param:
                        if item_param.count('=') == 1:
                            dot_string, value = item_param.split("=")
                            if "," in value:
                                value = value.split(",")
                            filter_params.append({
                                "dot_string": dot_string.strip(),
                                "value": value
                            })

            if len(filter_params) > 0:
                # Converte objetos em dot_string parametros
                def dot(d, dot_list):
                    options = ["$eq", "$gt", "$gte", "$in", "$lt", "$lte", "$ne", "$nin"]
                    keys = d.keys()
                    k = keys[0]
                    dot_list.append(k)

                    # Verificar se existe elementos filhos
                    if type(d[k]) == dict and d[k].keys() > 0:
                        _k = d[k].keys()[0]
                        if _k not in options:
                            return dot(d[k], dot_list)
                        else:
                            return d[k]
                    elif type(d) == dict and d.keys() > 0:
                        k = d.keys()[0]
                        return {"$eq": d[k]}

                # Converte string em objetos
                def put(d, dot_string, value):
                    options = ["eq", "gt", "gte", "in", "lt", "lte", "ne", "nin"]
                    if "." in dot_string:
                        key, rest = dot_string.split(".", 1)
                        if key not in d:
                            key = key if key not in options else '$' + key
                            d[key] = {}
                        put(d[key], rest, value)
                    else:
                        if dot_string not in options:
                            d[dot_string] = value
                        else:
                            dot_string = '$' + dot_string
                            d[dot_string] = value

                for item_filter in filter_params:
                    dot_list = []
                    d = {}
                    put(d, item_filter["dot_string"], item_filter["value"])
                    dot_result = dot(d, dot_list)
                    if dot_result != None:
                        result[(".".join(dot_list))] = dot_result

            return result

        if self.__filter != None:
            self.__filter = dot_query_operators(self.__filter)

    # --------------------------------
    # __limit
    # --------------------------------
    # Define o limite de documentos retornados
    #
    # ?limit=20
    #
    def limit(self, max_limit=0):
        # Condigura o limite default do serviço
        limit = error_limit = self.__default_limit

        # Verifica se o limit da ação foi definido
        if self.__limit > 0:
            # Recupera o limite da operação para validar
            limit = error_limit = self.__limit

            # Verifica se o limite da operação maior do que o maximo permitido pelo serviço
            if limit > self.__max_limit:

                # Caso seja, então registra o limite maximo permitido
                limit = error_max_limit = self.__max_limit

                # Verifica se um limite máximo foi definido pelo desenvolvedor do serviço
                if max_limit > 0 and limit > max_limit:
                    # Caso seja, então registra o limite maximo permitido
                    error_max_limit = max_limit
                    limit = max_limit

                self.add_error(85545454, "O limite máximo (" + str(
                    error_max_limit) + ") foi excedido. Limite solicitado (" + str(error_limit) + ").")

        self.__limit = limit

    def info(self):
        return {
            "owner": {
                "id": self.__token,
                "last_id": self.__token
            }
        }

    # --------------------------------
    # add_error
    # --------------------------------
    def add_error(self, code, message):
        self.__validate.add_error(code, message)

    # --------------------------------
    # add_error_required
    # --------------------------------
    def add_error_required(self, code, message):
        self.__validate.add_error_required(code, message)

    # --------------------------------
    # changed_time
    # --------------------------------
    def changed_time(self):
        self.__changed_time = str(datetime.datetime.utcnow())

    # --------------------------------
    # checksum
    # --------------------------------
    def checksum(self, data):
        md5 = hashlib.md5()
        md5.update(dumps(data))
        self.__checksum = md5.hexdigest()

    # --------------------------------
    # collection
    # --------------------------------
    def collection(self, collection):
        if collection != self.__service:
            list_names = collection.split("_")
            if list_names[0] != self.__service:
                list_names.insert(0, self.__service)
                collection = '_'.join(list_names)
        self.__collection = collection

    # --------------------------------
    # created_time
    # --------------------------------
    def created_time(self):
        self.__created_time = str(datetime.datetime.utcnow())

    # --------------------------------
    # ready
    # --------------------------------
    def ready(self):
        return self.__validate.ready()

    # --------------------------------
    # send_database
    # --------------------------------
    def send_database(self):
        # if self.__action == "save":
        return self.__send_save()

    def __send_save(self):
        message = {
            "action": self.__action,
            "collection": self.__collection,
            "connection_id": self.__connection_id,
            "cursor": self.__cursor,
            "data": self.__data,
            "database": self.__database,
            "error": self.__validate.error_to_list(),
            "fields": self.__fields,
            "filter": self.__filter,
            "limit": self.__limit,
            "sort": self.__sort,
            "service_id": self.__database_id,
            "source_id": self.__source_id,
            "info": self.info()
        }

        print "SEND SAVE", message

        # Montar mensagem que será enviada para o serviço de banco de dados
        return dumps(message)

    # --------------------------------
    # send_error
    # --------------------------------
    # Montar mensagem de erro para ser retornada ao servidor rest
    def send_error(self):
        result = {
            "connection_id": self.__connection_id,
            "success": False,
            "error": self.__validate.error_to_list()
        }
        return dumps({
            "connection_id": self.__connection_id,
            "message": result
        })

    # --------------------------------
    # validateService
    # --------------------------------
    def validateService(self, service, service_id):
        if self.__service != service or str(self.__service_id) != str(service_id):
            self.__validate.add_error("FALHA DE SERVICE", "Identificador do serviço nao valido:: " + self.__service + "!=" + service + "::" + str(
                self.__service_id) + "!=" + str(service_id))
            return False
        return True

    # --------------------------------
    # validateField
    # --------------------------------
    def validateField(self, item_scheme_name, item_scheme, value, action=None):
        return self.__validate.field(item_scheme_name, item_scheme, value, action)

    # ----------------------------------------------------------------
    #
    # Private Methods
    #
    # ----------------------------------------------------------------
    # --------------------------------
    # decode_service
    # --------------------------------
    def decode(self, message):
        # Prepara o pacote a ser enviado
        json = loads(message)

        print "DECODE", json

        self.body = json["body"]

        self.__action = json["action"]
        self.__connection_id = json["connection_id"]
        self.__cursor = json["cursor"]
        self.__fields = json["fields"]
        self.__filter = json["filter"]
        self.__limit = int(json["limit"])
        self.__info = json["info"]
        self.__offset = int(json["offset"])
        self.__service = json["service"]
        self.__service_id = int(json["service_id"])

        self.__sort = json["sort"]
        self.__source = self.__set_source(json["source"])
        self.__token = json["token"]
        self.__version = json["version"]

        # TODO:: Mudar regra de criação de nomes de banco de dados
        self.__database_id = self.__service_id + 20000
        # self.group = self.info["application"]["group"]
        self.__database = "db" + str(self.__database_id) + self.__service
        # self.info["database"] = {"name": self.database_name, "collection": None}

    # --------------------------------
    # __set_source
    # --------------------------------
    # Recupera o id do documento
    def __set_source(self, source):
        self.__source_id = None
        # Verificar se o source possui uma id
        # Pegar a id do source
        result = re.search(r"([\s\S]+)([0-9a-f]{24})$", source)
        if result:
            source = result.group(1) + ":id"
            self.__source_id = result.group(2)
        return source

    # --------------------------------
    # __set_source
    # --------------------------------

    def source(self):
        return self.__source
