#!/usr/bin/env python
# -*- coding: utf-8 -*-
from bson.int64 import Int64
from devdoo.datarest import DataRest
from devdoo.database import DataBase


class Service(object):
    def __init__(self, message, id, name, schemes, actions, custom_service=None):

        self.actions = actions
        self.id = id
        self.name = name
        self.schemes = schemes

        self.__database = DataBase()
        self.__datarest = DataRest(message, custom_service)

    # ----------------------------------------------------------------
    #
    # Public Methods
    #
    # ----------------------------------------------------------------
    # --------------------------------
    # action
    # --------------------------------
    def action(self):

        source = self.__datarest.source()

        # Verifica se o endereço da api existe
        if source in self.actions.keys():
            # Recupera o objeto de configuração da api
            service_api = self.actions[source]
            action = service_api["action"]

            if hasattr(self, action["database"]) and self.__datarest.validateService(self.name, self.id):
                method = getattr(self, action["database"])
                if action["database"] == "update":
                    method(action["collection"], action["scheme"], self.schemes)
                else:
                    method(action["collection"], self.__getScheme(action["scheme"], self.schemes), self.__get_max_limit(action))
            else:
                # TODO:: Melhorrar mensagem de erro para API INVALIDA E ACAO INVALIDA
                # TODO:: Informar na mensagem de erro que o id não  valido
                self.__datarest.add_error_required("AAA ACTION INVALIDA",
                                                   source + " :: action nao existe no servico '" + self.name + "' da versao::" + " " + str(
                                                       service_api["action"]))
        else:
            self.__datarest.add_error_required("BBB API INVALIDA", source + ":: endpoint da api nao existe no servico " + self.name + " da versao::")

    def __getScheme(self, scheme, list_schemes):
        result = {}
        for field_name in scheme:
            if field_name in list_schemes.keys():
                result[field_name] = list_schemes[field_name]
            else:
                self.__datarest.add_error_required("NOME DO CAMPO API não registrado", field_name)
        return result

    def __get_max_limit(self, action):
        if "max_limit" in action:
            return action["max_limit"]
        return None

    # --------------------------------
    # insert
    # --------------------------------
    # Prepara documento para ser inserido base de dados
    def insert(self, collection, list_scheme, max_limit=None):
        list_save_data = []

        # Define o tipo de ação que será executada no servidor de base de dados
        self.__datarest.action("insert")

        # Recupera o identificador da coleção
        self.__datarest.collection(collection)

        # Prepara a lista de campos válidos que poderão ser inseridos na base de dados
        # Recebe a lista de erros em campos recebidos no body
        # Recebe a lista de erros em campos definido pelo desenvolvedor de microserviço
        list_scheme_clean, body_field_errors, dev_fields_errors = self.__database.fields_save(self.__datarest.body, list_scheme)

        # Processa todos os campos registrados no schema para validar os valores e formatar
        for item_scheme_name in list_scheme_clean:
            # Recupera o equema para validar e formatar
            item_scheme = list_scheme_clean[item_scheme_name]

            # Recupera o valor que deverá ser validado e formatado
            field_value = item_scheme["value"]

            # Valida o campo e valor
            # Recebe o nome do campo que será inserido na base de dados e depois retornado
            # Recebe o valor que deverá ser insrido na base de dados
            field_name, field_value = self.__datarest.validateField(item_scheme_name, item_scheme, field_value, "custom_save")

            # Verifica se o campo não está com erro de validação
            if field_value != '__error__' and field_value != '__ERROR__':
                # Registra o campo na lista de dados que deverá ser enviada para o servidor de base de dados
                list_save_data.append({"field": field_name, "value": field_value})

        # Prepara s lista de campos que será enviada para o servidor de base de dados
        # Prepara a lista de campos que devrá ser retornada do servidor depois de salvar na base de dados
        list_save_data = self.__prepare_save_data_fields(list_save_data)

        # Registra os dados do documento
        self.__datarest.data(list_save_data)

        # Registra as datas de criação e alteração do documento
        self.__datarest.created_time()
        self.__datarest.changed_time()

        # Caso o campo nao esteja registrado no scheme, registra mensagem de error
        for item_field_name in body_field_errors:
            self.__datarest.add_error(125489, "O campo '" + item_field_name + "' nao existe.")

        # Caso o campo esteja em conflito no modo desenvolvedor de microserviço, registra mensagem de error
        for item_field_name in dev_fields_errors:
            self.__datarest.add_error_required(362548,
                                               "DEV ERROR O campo '" + item_field_name + "' nao pode ser utilizado por existir campos filhos.")

        # Verificar se tem algo para gravar no banco, caso contrário, registra mensagem de error
        if len(list_save_data) <= 0:
            self.__datarest.add_error_required(85545454, "Nao tem nada para gravar no banco de dados")

    # --------------------------------
    # find
    # --------------------------------
    def find(self, collection, list_scheme, max_limit=None):
        # Define o tipo de ação que será executada no servidor de base de dados
        self.__datarest.action("find")

        # Recupera o identificador da coleção
        self.__datarest.collection(collection)

        # Prepara a lista de campos válidos que poderão ser recuperados na base de dados
        # Recebe a lista de erros em campos definido pelo desenvolvedor de microserviço
        fields, dev_fields_errors = self.__database.fields_find(list_scheme)

        # Caso o campo esteja em conflito no modo desenvolvedor de microserviço, registra mensagem de error
        for item_field_name in dev_fields_errors:
            self.__datarest.add_error_required(362548,
                                               "DEV ERROR O campo '" + item_field_name + "' nao pode ser utilizado por existir campos filhos.")

        # Registra campos que devem ser retornados
        self.__datarest.fields(fields)
        self.__datarest.filter()
        self.__datarest.limit(max_limit)

        # self.__order() #TODO:: Implementar Order
        # self.__cursor() #TODO:: Implementar Cursos

    # --------------------------------
    # remove
    # --------------------------------
    def remove(self, collection, list_scheme, max_limit=None):
        self.__datarest.action("remove")
        self.__datarest.collection(collection)
        self.__filter()

    # --------------------------------
    # ready
    # --------------------------------
    def ready(self):
        return self.__datarest.ready()

    # --------------------------------
    # operator
    # --------------------------------
    # TODO:: implementar checksum de update
    def operator(self, list_scheme):

        list_scheme_clean, body_field_errors, dev_fields_errors = self.__database.fields_update(self.__datarest.body, list_scheme)

        self.__datarest.fields(fields_default)
        save_data = {}

        # Processa todos os campos registrados no schema para validar os valores e formatar
        for item_scheme in list_scheme_clean.values():
            field_name = item_scheme["field"]
            value = item_scheme["value"]

            # Valida o campo e valor correspondente
            value = self.__datarest.validateField(item_scheme, value, "custom_update")

            # Verifica se existe dados para gravar na base de dados
            if value != '__error__' and value != '__ERROR__':
                save_data[field_name] = value

        # Caso o campo nao esteja registrado no scheme, registra mensagem de error
        for item_field_name in body_field_errors:
            self.__datarest.add_error(125489, "O campo '" + item_field_name + "' nao existe.")

        for item_field_name in dev_fields_errors:
            self.__datarest.add_error_required(362548,
                                               "DEV ERROR O campo '" + item_field_name + "' nao pode ser utilizado por existir campos filhos.")

        return save_data

    # --------------------------------
    # send_database
    # --------------------------------
    def send_database(self):
        return self.__datarest.send_database()

    # --------------------------------
    # send_error
    # --------------------------------
    def send_error(self):
        return self.__datarest.send_error()

    # --------------------------------
    # update
    # --------------------------------
    def update(self, collection, scheme, list_scheme):
        self.__datarest.action("update")
        self.__datarest.collection(collection)

        save_data = {}
        for item_scheme in scheme:
            obj_item_scheme = self.__getScheme(scheme[item_scheme], list_scheme)
            obj_item_scheme = self.operator(obj_item_scheme)

            if len(obj_item_scheme.keys()) > 0:
                save_data[item_scheme] = obj_item_scheme

        self.__datarest.data(save_data)
        self.__filter()

        # Verificar se tem algo para gravar no banco
        if len(save_data.keys()) <= 0:
            self.__datarest.add_error_required(85545454, "Nao tem nada para ATUALIZAR no banco de dados")

    # ----------------------------------------------------------------
    #
    # Private Methods
    #
    # ----------------------------------------------------------------
    # --------------------------------
    # __fields
    # --------------------------------
    # Define a lista de campos a serem retornados em uma consulta
    #
    # ?fields=campoA,campoB,campoC
    #
    def __fields(self, fields_default):
        # TODO:: Implementar link para o schema
        # TODO:: Implementar controle de acesso para campos publicos/privados ("private": True)
        # TODO:: Implementar regras para retornar campos do _info (fields["_info"] = False)

        query = self.__datarest.query

        fields = {}
        for item_field in fields_default:
            fields[item_field] = True

        if 'fields' in query.keys():
            fields = {}
            query_fields = query["fields"]
            list_fields = query_fields.strip().split(',')

            # Marca campos para retorno
            for item_field in list_fields:
                if item_field in fields_default:
                    fields[item_field] = True
                else:
                    self.__datarest.add_error(3423423, "O campo (" + str(item_field) + ") nao existe.")

        fields["_id"] = False
        fields["id"] = True

        self.__datarest.fields(fields)





    # --------------------------------
    # __check_type
    # --------------------------------
    def __check_type(self, field_name, value, field_types):
        field_type = field_types[field_name]
        try:
            if field_type == 'number' or field_type == 'timestamp':
                return bson.int64(value)
            elif field_type == 'string':
                return str(value)
            elif field_type == 'boolean':
                return bool(value)
            elif field_type == 'array' or field_type == 'enum':
                return list(value)
            elif field_type == 'object':
                return dict(value)
            elif field_type == 'object_id':
                return bson.objectid.ObjectId(value)
            else:
                return value
        except:
            return value



    # --------------------------------
    # __prepare_save_data_fields
    # --------------------------------
    # Prepara lista de campos que podem ser incluidos no documento
    # Prepara e registra a lista de campos que deverá ser retornada do servidor após a inclusão do documento na base de dados
    def __prepare_save_data_fields(self, list_data):
        result = {}
        fields = {}

        for item in list_data:
            # Verifica se o nome do campo existe na lista que será incluida no documento
            if (item["field"]) in result.keys():
                # Caso não esteja na lista então o campo é incluido
                # Prepara objetos filhos no padrão permitido paraga registrar na base de dados
                result[(item["field"])] = self.__save_data_join(result[(item["field"])], item["value"])
            else:
                # Registra o campo e valor na lista de campos válida do documento
                result[(item["field"])] = item["value"]

            # Adiciona o campo na lista de campos que devem ser retornados do servidor
            fields[(item["field"])] = True

        # Registra campos que devem ser retornados
        self.__datarest.fields(fields)

        return result

    # --------------------------------
    # __save_data_join
    # --------------------------------
    # Função recursiva utilizada para tratar conflito de elementos filho no momento de gravar na base de dados
    #
    # data - Objeto contento dados existentes do mesmo grupo ou um objeto vazio para receber novos elementos
    # value - Dados que deverão ser inseridos no objeto
    def __save_data_join(self, data, value):

        # Pega cada elemento do objeto valor
        for item in value:
            # Verifica se valor é um objeto
            # Verifica se o elemento filho valor é um objeto
            # Verifica se o elemento filho existe no objeto de dados
            if type(value) == dict and type(value[item]) == dict and type(data) == dict and item in data.keys():
                # Adiciona novo elemento no objeto de dados com o resultado da interação nos objetos filhos do objeto valor
                data[item] = self.__save_data_join(data[item], value[item])

            # Verifica se valor é um objeto
            # Verifica se o elemento filho valor é um objeto
            elif type(value) == dict and type(data) == dict:
                # Adiciona o objeto valor como novo elemento no objeto de dados
                data[item] = value[item]

            # Verifica se o objeto dados é um objeto
            elif type(data) == dict:
                # Adiciona o valor como novo elemento no objeto de dados
                data[item] = value
        return data
